"""Migration command for dbt CI."""
import sys
import logging
from argparse import Namespace
from typing import cast
import click
from src.cache import CacheManager
from src.graph.dependency_graph import DbtGraph
from src.logging import print_exception
from src.schema import MigrationMap, SupportedConnectors
from src.connectors import get_connector
from src.utilities.paths import get_profile
from src.graph.graph_utils import (
    filter_node_ids_by_multiple_types,
    get_node_ids_from_structured_nodes,
    get_nodes
)

logger = logging.getLogger(__name__)

def migration(args: Namespace):
    """
    Perform a migration operation.

    This function serves as a placeholder for the migration command.
    It currently does not implement any functionality.

    Args:
        args: Namespace object containing command-line arguments.
    """
    try:
        click.secho("DBT CI Migration", fg="green", bold=True)
        logger.debug(f"Running with the following arguments: {args}")
        cache = CacheManager(args)
        cache.start_report("migrate", args)
        connector_type = cast(SupportedConnectors, get_profile(args)["type"])
        connector = get_connector(connector_type)

        if connector is None:
            logger.error(f"Connector '{connector_type}' does not support migration strategy, which is required for migration command.")
            sys.exit(1)

        # Setup migraiton connector
        migration_connector = connector.get("strategies").get("migration", None)
        if migration_connector is None:
            logger.error(f"Connector '{connector_type}' does not have a migration strategy implemented, which is required for migration command.")
            sys.exit(1)

        # Determine if there has been a change in clustering or partitioning configuration
        migration_map = generate_migration_map(args, cache)

        if len(migration_map["nodes"]) == 0:
            logger.info("No partitioning changes detected between target and reference graphs.")
            sys.exit(0)
        else:
            logger.info(f"Detected {len(migration_map['nodes'])} models with partitioning changes.")
            logger.info("\n------------------------------------------------------")
            for node_id, node_info in migration_map["nodes"].items():
                logger.info(f"Model: {node_id}")
                logger.info(f"  - Table ID: {node_info['table_id']}")
                logger.info(f"  - Old Partitioning: {node_info['old_partitioning']}")
                logger.info(f"  - New Partitioning: {node_info['new_partitioning']}")
            logger.info("------------------------------------------------------\n")

        if getattr(args, "dry_run", False):
            logger.info("Dry run mode enabled - no changes will be applied.")
            sys.exit(0)
        
        # Apply partitioning changes
        migration_connector(migration_map, args)
        cache.update_report("migrate", "completed", comment=str(list(migration_map["nodes"].keys())))
        logger.info("Migration completed successfully.")
    except Exception as e:
        cache = CacheManager(args)
        cache.update_report("migrate", "failed", comment=str(e))
        print_exception(e)
        sys.exit(1)

def generate_migration_map(
    args: Namespace,
    cache: CacheManager,
) -> MigrationMap:
    """Compare partitioning configurations between target and reference nodes and return a migration map."""
    target_graph = DbtGraph(args)
    reference_graph = DbtGraph(args, is_reference=True)

    # Look for cache
    prev_cache = cache.get_cache()
    if prev_cache is None: # Should we exit here instead of compiling?
        logger.error("No cache found, please run 'model-lineage init' first to generate the necessary manifest files and cache for comparison.")
        sys.exit(1)
    logger.info("Cache successfully found - using cached state for comparison")

    modified_nodes_dict = {
        "modified_nodes": get_node_ids_from_structured_nodes(cache.get_cache().get("modified_nodes", None)) or [],
    }

    selected_nodes = filter_node_ids_by_multiple_types(
        dependency_graph=target_graph.to_dict(), 
        node_ids=modified_nodes_dict["modified_nodes"], 
        node_types=["model"]
    )
    if len(selected_nodes) == 0:
        logger.info("No modified models found in cache, skipping...")
        sys.exit(0)

    target_nodes = get_nodes(target_graph.to_dict(), selected_nodes)
    reference_nodes = get_nodes(reference_graph.to_dict(), selected_nodes)

    migration_map: MigrationMap = {
        "connector": get_profile(args)["type"],
        "nodes": {}
    }

    for node_id, node_metadata in target_nodes.items():
        node_config = node_metadata.get("config", {})
        target_partitioning_config = node_config.get("partition_by", None)
        reference_partitioning_config = reference_nodes.get(node_id, {}).get("config", {}).get("partition_by", None)
        # Skip non-incremental models since partitioning changes only apply to incremental models in BigQuery
        if node_config.get("materialized", None) != "incremental":
            continue
        
        # Check if partitioning has been changed
        # Dict comparison is order-insensitive since Python 3.7+
        if target_partitioning_config != reference_partitioning_config:
            migration_map["nodes"][node_id] = {
                "table_id": f"{node_metadata.get('database', '')}.{node_metadata.get('schema', '')}.{node_metadata.get('name', '')}",
                "compiled_code": node_metadata.get("compiled_code", None),
                "old_partitioning": reference_partitioning_config,
                "new_partitioning": target_partitioning_config
            }

    return migration_map


import logging
import os
import sys
import yaml
import json
from pathlib import Path
from argparse import Namespace
from src.schema import DBTManifest, DBTProfile

logger = logging.getLogger(__name__)

def get_dir_name_from_path(path: str) -> str | None:
    """Get the directory name from a given path."""
    if os.path.isdir(path):
        return os.path.basename(os.path.normpath(path))
    elif os.path.isfile(path):
        return os.path.basename(os.path.dirname(path))
    elif not os.path.exists(path):
        logger.error(f"Path '{path}' does not exist.")
        sys.exit(1)

def get_manifest_file(dbt_project_dir: str) -> DBTManifest:
    """
    Get the path to the manifest.json file in the target directory.
    Raises a FileNotFoundError if the file does not exist.
    """

    file = os.path.join(dbt_project_dir, "target/manifest.json")
    if not os.path.isfile(file):
        raise FileNotFoundError(f"manifest.json not found in {dbt_project_dir}")
    with open(file, 'r', encoding='utf-8') as f:
        return json.load(f)
    
def get_manifest_file_full_path(file_path: str) -> DBTManifest | None:
    """
    Get the manifest.json file from the specified path.
    Raises a FileNotFoundError if the file does not exist.
    Uses pydantic to confirm the structure of the manifest file.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"manifest.json not found at {file_path}")
    with open(file_path, 'r', encoding='utf-8') as f:
        item = json.load(f)
        if isinstance(item, DBTManifest):
            return item

def get_reference_manifest_file(reference_state_dir: str) -> DBTManifest:
    """
    Get the reference manifest.json file from the state directory.
    Raises a FileNotFoundError if the file does not exist.
    """
    file = os.path.join(reference_state_dir, "manifest.json")
    if os.path.isfile(reference_state_dir):
        # If the provided reference_state_dir is actually a file, use it directly
        file = reference_state_dir

    if not os.path.isfile(file):
        raise FileNotFoundError(f"manifest.json not found in {reference_state_dir}")
    with open(file, 'r', encoding='utf-8') as f:
        return json.load(f)

def get_dbt_project_file(dbt_project_dir: str) -> dict:
    """
    Get the path to the dbt_project.yml file in the specified directory.
    Raises a FileNotFoundError if the file does not exist.
    """
    file = os.path.join(dbt_project_dir, "dbt_project.yml")
    if not os.path.isfile(file):
        raise FileNotFoundError(f"dbt_project.yml not found in {dbt_project_dir}")
    
    with open(file, 'r', encoding='utf-8') as f:
        return yaml.safe_load(f)

def get_profiles_file(dbt_project_dir: str, profiles_dir: str | None = None):
    """
    Get the path to the profiles.yml file in the specified directory or the default locations.
    Raises a FileNotFoundError if the file does not exist in any of the expected locations
    """
    if profiles_dir:
        file = os.path.join(profiles_dir, "profiles.yml")
        if not os.path.isfile(file):
            raise FileNotFoundError(f"profiles.yml not found in {profiles_dir}")
        
        with open(file, 'r', encoding='utf-8') as f:
            return yaml.safe_load(f)
    else:
        # Check for profiles.yml in the dbt project directory
        file = os.path.join(dbt_project_dir, "profiles.yml")
        if os.path.isfile(file):
            with open(file, 'r', encoding='utf-8') as f:
                return yaml.safe_load(f)
        
        # Check for profiles.yml in the user's home directory
        home_dir = os.path.expanduser("~")
        file = os.path.join(home_dir, ".dbt/profiles.yml")
        if os.path.isfile(file):
            with open(file, 'r', encoding='utf-8') as f:
                return yaml.safe_load(f)

        raise FileNotFoundError("profiles.yml not found in the specified profiles directory, dbt project directory, or ~/.dbt/")

## New methods, review them
def get_dbt_project_config(dbt_root_path: str) -> dict:
    """Get the DBT project configuration from the dbt_project.yml file.
    Raises a FileNotFoundError if the dbt_project.yml file does not exist in the specified directory.
    """
    dbt_project_path = get_dbt_project_path(dbt_root_path)
    if dbt_project_path is None:
        logger.error("DBT project path could not be resolved.")
        sys.exit(1)
    dbt_project_path_yaml_path = os.path.join(dbt_project_path, "dbt_project.yml")
    if not os.path.exists(dbt_project_path_yaml_path):
        logger.error(f"dbt_project.yml not found at {dbt_project_path_yaml_path}")
        sys.exit(1)
    with open(dbt_project_path_yaml_path, 'r', encoding='utf-8') as file:
        project_config = yaml.safe_load(file)

    return project_config

def get_dbt_project_path(dbt_project_path: str) -> Path:
    """
    Get the DBT project root directory path.

    Args:
        dbt_project_path: Path to the DBT project directory

    Returns:
        Path: Absolute path to DBT project directory
    """
    if not dbt_project_path:
        logger.error("DBT project path not provided.")
        return sys.exit(1)

    project_path = Path(dbt_project_path)

    # Convert relative to absolute
    if not project_path.is_absolute():
        project_path = Path.cwd() / project_path

    return project_path.resolve()

def get_dbt_profiles_config(dbt_root_path: str) -> dict:
    """Get the DBT profiles configuration from the profiles.yml file."""
    dbt_project_path = get_dbt_project_path(dbt_root_path)
    if dbt_project_path is None:
        logger.error("DBT project path could not be resolved.")
        sys.exit(1)

    profiles_path = os.path.join(dbt_project_path, "profiles.yml")
    if not os.path.exists(profiles_path):
        logger.error(f"profiles.yml not found at {profiles_path}")
        sys.exit(1)

    with open(profiles_path, 'r', encoding='utf-8') as file:
        profiles = yaml.safe_load(file)

    return profiles

def get_profile(args: Namespace) -> DBTProfile:
    """Get the DBT profile output configuration based on the provided arguments."""
    dbt_root_path = getattr(args, "dbt_project_dir", None)
    if dbt_root_path is None:
        logger.error("DBT project path not specified in arguments. Please pass the path using --dbt-project-dir <path>.")
        sys.exit(1)

    dbt_project_file = get_dbt_project_config(dbt_root_path)
    profiles = get_dbt_profiles_config(dbt_root_path)
    dbt_profile = dbt_project_file.get("profile", None)
    target = getattr(args, "target", None) or profiles.get(dbt_profile, {}).get("target", None)
    profile_name = getattr(args, "profile", "default")

    if target is None:
        logger.error("Target environment not specified in arguments. Please pass the target using --target <target>.")
        sys.exit(1)
    
    # Get the default profile (assuming 'default' is the profile name)
    if profile_name not in profiles:
        logger.error(f"Profile '{profile_name}' not found in profiles.yml")
        sys.exit(1)
    
    profile = profiles[profile_name]
    
    # Get the output configuration for the target
    if "outputs" not in profile or target not in profile["outputs"]:
        logger.error(f"Target '{target}' not found in profile '{profile_name}'")
        sys.exit(1)
    
    output_config: DBTProfile = profile["outputs"][target]

    return output_config

def get_absolute_path(path: str) -> str:
    """Return an absolute version of the given path."""
    return str(Path(path).resolve())



# Shared options for all commands
import yaml
import os
import re
import click
import logging

logger = logging.getLogger(__name__)

def _load_config_callback(ctx, param, value):
    """
    Eager callback that reads the dbt-ci YAML config file and injects its values
    into os.environ before Click resolves the remaining options.

    Keys in the config file should match Click envvar names (e.g. DBT_RUNNER).
    Actual shell environment variables always take precedence (setdefault).

    Values of the form ${VAR_NAME} are resolved from the current environment.

    Example config (dbt-ci.config.yaml):
        DBT_RUNNER: docker
        DBT_DOCKER_IMAGE: europe-west1-docker.pkg.dev/my-project/dbt
        DBT_PROJECT_DIR: dbt
        DBT_DOCKER_ENV: "DBT_PROFILES_DIR=/dbt,GOOGLE_APPLICATION_CREDENTIALS=${GOOGLE_APPLICATION_CREDENTIALS}"
    """

    def _resolve(val: str) -> str:
        """Replace ${VAR} references with their current environment values."""
        def replacer(match):
            var_name = match.group(1)
            return os.environ.get(var_name, match.group(0))  # keep original if not found
        return re.sub(r'\$\{([^}]+)\}', replacer, val)

    if not value:
        return value
    try:
        if os.path.exists(value) is False:
            return value

        logger.info(f"Configuration file found. Loading configuration from {value}")

        with open(value, "r", encoding="utf-8") as f:
            config = yaml.safe_load(f) or {}
        for key, val in config.items():
            if val is not None:
                if isinstance(val, list):
                    resolved = ",".join(_resolve(str(item)) for item in val)
                else:
                    resolved = _resolve(str(val))
                os.environ.setdefault(str(key), resolved)
    except Exception:
        logger.debug(f"No valid config file found at {value} or error reading file - proceeding with environment variables and defaults.")
        pass  # Config file issues should not crash the CLI
    return value

@click.group()
@click.version_option(version='0.1.0', prog_name='dbt-ci')
def cli():
    """dbt CI Tool - Intelligent CI for dbt projects
    
    Detect, run, and test only what changed based on state comparison.
    
    Visit https://datablock.dev for more information.
    """
    pass

def parse_multiple_option(ctx, param, value):
    """
    Parse multiple values for options that accept repeated flags.

    - Multiple CLI flags (--docker-env A --docker-env B): already a tuple, pass through.
    - Single string from env var with comma or newline separation: split and return tuple.

    Examples:
        DBT_DOCKER_ENV="KEY1=val1,KEY2=val2"          -> ('KEY1=val1', 'KEY2=val2')
        DBT_DOCKER_ENV=$'KEY1=val1\nKEY2=val2'         -> ('KEY1=val1', 'KEY2=val2')
        --docker-env KEY1=val1 --docker-env KEY2=val2  -> ('KEY1=val1', 'KEY2=val2') (unchanged)
    """
    if not value:
        return value
    if len(value) == 1 and (',' in value[0] or '\n' in value[0]):
        return tuple(p.strip() for p in value[0].replace('\n', ',').split(',') if p.strip())
    return value

def common_options(f):
    """Decorator to add common options to all commands"""
    # Dynamic package options
    # Config file — loaded eagerly so its env var values are injected before
    # Click resolves the remaining options
    f = click.option(
        "--config", "-c",
        envvar=['DBT_CONFIG'],
        default="dbt-ci.config.yaml",
        is_eager=True,
        callback=_load_config_callback,
        help="Path to dbt-ci configuration file (YAML). Keys must be env var names (e.g. DBT_RUNNER). Shell env vars always take precedence."
    )(f)
    f = click.option(
        "--dbt-version",
        envvar=['DBT_VERSION'],
        default=None,
        help="Specify a dbt version to use for this command"
    )(f)
    f = click.option(
        "--adapter", "-a",
        envvar=['DBT_ADAPTER'],
        type=str,
        default=None,
        help="Specify the dbt adapter to use (e.g., 'postgres', 'snowflake')"
    )(f)
    # Main commands
    f = click.option(
        '--dbt-project-dir', 
        envvar=['DBT_PROJECT_DIR'],
        default='.',
        help='Path to the dbt project directory (default: current directory)'
    )(f)
    f = click.option(
        '--profiles-dir', 
        envvar=['DBT_PROFILES_DIR'],
        default=None,
        help='Path to the directory containing the dbt profiles.yml file'
    )(f)
    f = click.option(
        '--reference-state', '--state',
        envvar=['DBT_STATE'],
        default=None,
        type=str,
        help='Path to the reference manifest.json directory (local path where state will be downloaded)'
    )(f)
    f = click.option(
        '--target', 
        '-t',
        envvar=['DBT_TARGET'],
        default=None,
        help='The dbt target to use (defaults to what is defined in profiles.yml)'
    )(f)
    f = click.option(
        '--vars', 
        '-v',
        envvar=['DBT_VARS'],
        default='',
        help='A YAML string or path to YAML file containing variables to pass to dbt'
    )(f)
    f = click.option(
        "--defer",
        envvar=['DBT_DEFER'],
        is_flag=True,
        default=False,
        help="Use dbt's --defer flag to defer to the state of the production manifest (only applicable to run and test commands)"
    )(f)
    f = click.option(
        '--runner', 
        '-r', 
        envvar=['DBT_RUNNER'],
        type=click.Choice(['local', 'docker', 'bash', 'dbt']),
        default='dbt',
        help='Runner to use for executing dbt commands'
    )(f)
    f = click.option(
        '--entrypoint',
        envvar=['DBT_ENTRYPOINT'],
        default='dbt',
        help='Command entrypoint for dbt (default: dbt)'
    )(f)
    f = click.option(
        '--dry-run',
        envvar=['DBT_DRY_RUN'],
        is_flag=True,
        default=False,
        help='Print commands without executing them'
    )(f)
    f = click.option(
        '--log-level',
        envvar=['DBT_LOG_LEVEL'],
        type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], case_sensitive=False),
        default='INFO',
        callback=lambda ctx, param, value: value.upper() if value else value,
        help='Logging level'
    )(f)
    f = click.option(
        "--slack-webhook",
        "--slack-webhook-url",
        envvar=['SLACK_WEBHOOK', 'SLACK_WEBHOOK_URL'],
        default=None,
        type=str,
        help="Slack webhook URL for notifications (optional)"
    )(f)
    # Docker options
    f = click.option(
        '--docker-image',
        envvar=['DBT_DOCKER_IMAGE'],
        default='ghcr.io/dbt-labs/dbt-core:latest',
        help='Docker image to use'
    )(f)
    f = click.option(
        '--docker-platform',
        envvar=['DBT_DOCKER_PLATFORM'],
        default=None,
        help='Platform for Docker (e.g., linux/amd64)'
    )(f)
    f = click.option(
        '--docker-volumes',
        envvar=['DBT_DOCKER_VOLUMES'],
        multiple=True,
        callback=parse_multiple_option,
        help='Additional volume mounts (format: host:container). Repeat flag for multiple volumes: --docker-volumes /path1:/path1 --docker-volumes /path2:/path2. Via env var, use comma or newline separation: DBT_DOCKER_VOLUMES="/p1:/p1,/p2:/p2"'
    )(f)
    f = click.option(
        '--docker-env',
        envvar=['DBT_DOCKER_ENV'],
        multiple=True,
        callback=parse_multiple_option,
        help='Environment variables (format: KEY=VALUE). Repeat flag for multiple vars: --docker-env VAR1=val1 --docker-env VAR2=val2. Via env var, use comma or newline separation: DBT_DOCKER_ENV="KEY1=val1,KEY2=val2"'
    )(f)
    f = click.option(
        '--docker-network',
        envvar=['DBT_DOCKER_NETWORK'],
        default='host',
        help='Docker network mode'
    )(f)
    f = click.option(
        '--docker-user',
        envvar=['DBT_DOCKER_USER'],
        default=None,
        help='User to run as inside container'
    )(f)
    f = click.option(
        '--docker-args',
        envvar=['DBT_DOCKER_ARGS'],
        default='',
        help='Additional docker run arguments'
    )(f)
    # Bash options
    f = click.option(
        '--shell-path', 
        '--bash-path',
        envvar=['DBT_SHELL_PATH'],
        default='/bin/bash',
        help='Path to shell executable for bash runner'
    )(f)
    f = click.option(
        "--quiet", "-q",
        envvar=['DBT_QUIET'],
        is_flag=True,
        default=False,
        help="Run in quiet mode with minimal output"
    )(f)

    return f

# Import command modules to register them with the cli group
import src.commands.init.cli       # noqa: E402, F401
import src.commands.run.cli        # noqa: E402, F401
import src.commands.ephemeral.cli  # noqa: E402, F401
import src.commands.delete.cli     # noqa: E402, F401
import src.commands.migration.cli  # noqa: E402, F401
import src.commands.finalize.cli   # noqa: E402, F401



"""Unit tests for the run command."""
import pytest
from unittest.mock import MagicMock, patch, call
from argparse import Namespace
from src.commands.run.index import run as index, run_with_mode


class TestRunCommand:
    """Test the run command."""
    
    @patch('src.commands.run.index.CacheManager')
    @patch('src.commands.run.index.DbtGraph')
    @patch('src.commands.run.index.logger.error')
    @patch('src.commands.run.index.click.secho')
    def test_run_no_cache(
        self,
        mock_secho,
        mock_logger_error,
        mock_graph,
        mock_cache
    ):
        """Test run command exits when no cache is found."""
        # Setup mocks
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = None
        mock_cache.return_value = mock_cache_instance
        # Run command
        args = Namespace(
            dbt_project_dir='/dbt',
            dry_run=False
        )
        index(args)
        
        # Verify message was logged
        mock_logger_error.assert_called_with(
            "No cache found, please run 'dbt-ci init' first to generate the necessary manifest files and cache for comparison."
        )
    
    @patch('src.commands.run.index.CacheManager')
    @patch('src.commands.run.index.DbtGraph')
    @patch('src.commands.run.index.get_node_ids_from_structured_nodes')
    @patch('src.commands.run.index.run_with_mode')
    @patch('src.commands.run.index.logger.info')
    @patch('src.commands.run.index.click.secho')
    def test_run_no_modified_nodes(
        self,
        mock_secho,
        mock_logger_info,
        mock_run_mode,
        mock_get_nodes,
        mock_graph,
        mock_cache
    ):
        """Test run command exits when no modified nodes are found."""
        # Setup mocks
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": None,
            "new_nodes": None,
            "deleted_nodes": None
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_nodes.return_value = []
        
        # Run command - expect it to exit with code 0
        args = Namespace(
            dbt_project_dir='/dbt',
            dry_run=False
        )
        with pytest.raises(SystemExit) as exc_info:
            index(args)
        
        # Verify it exited with code 0
        assert exc_info.value.code == 0
        
        # Verify appropriate message was logged
        mock_logger_info.assert_any_call("No modified, new, or deleted nodes found in cache, skipping...")
    
    @patch('src.commands.run.index.CacheManager')
    @patch('src.commands.run.index.DbtGraph')
    @patch('src.commands.run.index.get_node_ids_from_structured_nodes')
    @patch('src.commands.run.index.run_with_mode')
    @patch('src.commands.run.index.click.secho')
    @patch('src.commands.run.index.click.echo')
    def test_run_with_modified_nodes(
        self,
        mock_echo,
        mock_secho,
        mock_run_mode,
        mock_get_nodes,
        mock_graph,
        mock_cache
    ):
        """Test run command with modified nodes."""
        # Setup mocks
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": {"model": {"model.project.model1": {}}},
            "new_nodes": {"model": {"model.project.model2": {}}},
            "deleted_nodes": None
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_nodes.side_effect = [
            ['model.project.model1'],  # modified
            ['model.project.model2'],  # new
            []  # deleted
        ]
        
        mock_graph_instance = MagicMock()
        mock_graph_instance.to_dict.return_value = {}
        mock_graph.return_value = mock_graph_instance
        
        # Run command
        args = Namespace(
            dbt_project_dir='/dbt',
            nodes='models',
            dry_run=False
        )
        index(args)
        
        # Verify run_with_mode was called
        mock_run_mode.assert_called_once()
        call_args = mock_run_mode.call_args
        assert call_args[1]['mode'] == 'models'
    
    @patch('src.commands.run.index.click.echo')
    @patch('src.commands.run.index.sys.exit')
    def test_run_error_handling(
        self,
        mock_exit,
        mock_echo
    ):
        """Test run command error handling."""
        # Run with invalid kwargs to trigger exception
        args = Namespace(
            dbt_project_dir='/dbt',
            dry_run=False
        )
        index(args)
        
        # Verify error was handled
        mock_exit.assert_called_once_with(1)


class TestRunWithMode:
    """Test the run_with_mode helper function."""
    
    @patch('src.commands.run.index.run_dbt_command')
    @patch('src.commands.run.index.get_downstream_dependencies')
    @patch('src.commands.run.index.filter_node_ids_by_type')
    @patch('src.commands.run.index.click.echo')
    def test_run_with_mode_all(
        self,
        mock_echo,
        mock_filter,
        mock_downstream,
        mock_run_cmd
    ):
        """Test run_with_mode with 'all' mode."""
        # Setup
        variables = Namespace(dry_run=False)
        target_graph = MagicMock()
        target_graph.to_dict.return_value = {
            "model": {},
            "seed": {},
            "snapshot": {},
            "test": {}
        }
        
        modified_nodes_dict = {
            "modified_nodes": ["model.project.model1"],
            "new_nodes": ["model.project.model2"],
            "deleted_nodes": []
        }
        modified_nodes = ["model.project.model1", "model.project.model2"]
        
        mock_downstream.return_value = []
        mock_filter.return_value = ["model.project.model1", "model.project.model2"]
        
        mock_result = MagicMock()
        mock_result.returncode = 0
        mock_run_cmd.return_value = mock_result
        
        # Run
        run_with_mode('all', variables, target_graph, modified_nodes_dict, modified_nodes)
        
        # Verify all command types were attempted
        # Should run: seed, run, test, snapshot
        assert mock_run_cmd.call_count >= 1
    
    @patch('src.commands.run.index.run_dbt_command')
    @patch('src.commands.run.index.get_downstream_dependencies')
    @patch('src.commands.run.index.filter_node_ids_by_type')
    @patch('src.commands.run.index.click.echo')
    def test_run_with_mode_models_only(
        self,
        mock_echo,
        mock_filter,
        mock_downstream,
        mock_run_cmd
    ):
        """Test run_with_mode with 'models' mode."""
        # Setup
        variables = Namespace(dry_run=False)
        target_graph = MagicMock()
        target_graph.to_dict.return_value = {"model": {}}
        
        modified_nodes_dict = {
            "modified_nodes": ["model.project.model1"],
            "new_nodes": [],
            "deleted_nodes": []
        }
        modified_nodes = ["model.project.model1"]
        
        mock_downstream.return_value = ["model.project.model3"]
        mock_filter.return_value = ["model.project.model1", "model.project.model3"]
        
        mock_result = MagicMock()
        mock_result.returncode = 0
        mock_run_cmd.return_value = mock_result
        
        # Run
        run_with_mode('models', variables, target_graph, modified_nodes_dict, modified_nodes)
        
        # Verify run command was called
        mock_run_cmd.assert_called_once()
    
    @patch('src.commands.run.index.get_downstream_dependencies')
    @patch('src.commands.run.index.filter_node_ids_by_type')
    @patch('src.commands.run.index.logger.info')
    def test_run_with_mode_no_nodes(
        self,
        mock_logger_info,
        mock_filter,
        mock_downstream
    ):
        """Test run_with_mode when no nodes match the filter."""
        # Setup
        variables = Namespace(dry_run=False)
        target_graph = MagicMock()
        target_graph.to_dict.return_value = {}
        
        modified_nodes_dict = {
            "modified_nodes": [],
            "new_nodes": [],
            "deleted_nodes": []
        }
        modified_nodes = []
        
        mock_downstream.return_value = []
        mock_filter.return_value = []
        
        # Run
        run_with_mode('models', variables, target_graph, modified_nodes_dict, modified_nodes)
        
        # Verify no models message was logged
        mock_logger_info.assert_any_call("No models to run")
    
    @patch('src.commands.run.index.run_dbt_command')
    @patch('src.commands.run.index.get_downstream_dependencies')
    @patch('src.commands.run.index.filter_node_ids_by_type')
    @patch('src.commands.run.index.logger.info')
    def test_run_with_mode_dry_run(
        self,
        mock_logger_info,
        mock_filter,
        mock_downstream,
        mock_run_cmd
    ):
        """Test run_with_mode in dry run mode."""
        # Setup
        variables = Namespace(dry_run=True)
        target_graph = MagicMock()
        target_graph.to_dict.return_value = {"model": {}}
        
        modified_nodes_dict = {
            "modified_nodes": ["model.project.model1"],
            "new_nodes": [],
            "deleted_nodes": []
        }
        modified_nodes = ["model.project.model1"]
        
        mock_downstream.return_value = []
        mock_filter.return_value = ["model.project.model1"]
        
        # Run
        run_with_mode('models', variables, target_graph, modified_nodes_dict, modified_nodes)
        
        # Verify dbt command was NOT executed
        mock_run_cmd.assert_not_called()
        
        # Verify dry run message was logged
        mock_logger_info.assert_any_call("DRY RUN: Command would be executed")



"""Utility functions for multithreading."""
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Any

def run_multithreaded(
    func_list: list[Callable[..., Any]],
    threads: int = 4,
    exit_on_exception: bool = False
) -> list[Any]:
    """
    Run a list of functions concurrently using multithreading.

    Args:
        func_list: List of functions that may take any arguments.
        threads: Number of worker threads.
        exit_on_exception: If True, stop on first exception.

    Returns:
        List of results in the same order as func_list.
        Exceptions are stored in the list if exit_on_exception is False.
    """
    results = []

    with ThreadPoolExecutor(max_workers=threads) as executor:
        try:
            for result in executor.map(lambda f: f(), func_list):
                results.append(result)
        except Exception as e:
            if exit_on_exception:
                print(f"Exception: {e}", file=sys.stderr)
                sys.exit(1)
            else:
                results.append(e)

    return results



"""
DbtGraph class for representing dbt dependencies and lineage.
This module defines the DbtGraph class, which encapsulates the dependency graph of dbt resources
"""

import json
from argparse import Namespace
from typing import cast
from src.cache import CacheManager
from src.schema import DBTManifest, DependencyGraph
from src.graph.parser import generate_dependency_graph
from src.utilities.paths import get_manifest_file, get_reference_manifest_file

class DbtGraph:
    """
    Structured representation of dbt dependencies for lineage analysis.
    This class encapsulates the dependency graph of dbt resources, providing methods to access and analyze the relationships between models, macros, sources, seeds, snapshots, tests, and exposures. It also includes functionality to determine which nodes have been modified based on the state comparison between the production manifest and the target manifest.
    The DbtGraph class is initialized with the command-line arguments provided by the user, which include paths to the dbt project, production manifest, profiles directory, and other configuration options. It generates the dependency graph using the provided manifest files and allows users to query for modified nodes and their dependencies.

    Args:
        args (Namespace): Command-line arguments parsed
        is_production (bool, optional): Flag indicating whether to use the production state for comparison. Defaults to False.

    Returns:
        DbtGraph: An instance of the DbtGraph class containing the dependency graph and related
    """
    def __init__(self, args: Namespace, is_reference: bool = False):
        cache = CacheManager(args)
        self.args = args
        self.is_reference = is_reference
        for key, value in self.args.__dict__.items():
            setattr(self, key, value)
        
        # Keep paths as provided by user (relative or absolute)
        if getattr(self.args, "command", None) == "init":
            self.dbt_project_dir = self.args.dbt_project_dir
            self.dependency_graph = generate_dependency_graph(
                manifest_file=get_manifest_file(self.dbt_project_dir) if not self.is_reference 
                else get_reference_manifest_file(self.args.reference_state),
            )
        else: # We retrive manifest file from cache
            manifest_file = cache.get_cache("target_manifest.json" if not self.is_reference else "reference_manifest.json")
            self.dependency_graph = generate_dependency_graph(cast(DBTManifest, manifest_file))

    def to_dict(self) -> DependencyGraph:
        """Convert the DependencyGraph instance to a dictionary."""
        return self.dependency_graph

    def to_json(self, destination_path: str = "./dependency_graph.json") -> None:
        """Convert the DependencyGraph instance to a JSON string."""
        with open(destination_path, 'w', encoding='utf-8') as file:
            json.dump(
                obj=self.dependency_graph, 
                fp=file,
                indent=4,
                default=lambda o: list(o) if isinstance(o, set) else o
            )



"""
dbt-ci commands module

This module contains all CLI commands for dbt-ci.
"""

from src.commands.run.index import run
from src.commands.ephemeral.index import ephemeral
from src.commands.init.index import init
from src.commands.delete.index import delete
from src.commands.finalize.index import finalize
from src.commands.migration.index import migration

__all__ = ['run', 'ephemeral', 'init', 'delete', "finalize", "migration"]



import pytest
from typing import cast
from src.schema import RunnerConfig


@pytest.fixture
def mock_runner_config():
    """
    Factory fixture that returns a RunnerConfig dict.
    Allows overriding default values by passing a dict.
    """

    default_config: RunnerConfig = {
        'runner': 'docker',
        'dbt_project_dir': '/workspace/dbt',
        'reference_state': '/workspace/dbt/.dbtstate',
        'profiles_dir': '/workspace/dbt',
        'target': 'dev',
        'vars': '',
        'entrypoint': 'dbt',
        'dry_run': False,
        'quiet': False,
        'log_level': 'INFO',
        'docker_image': 'ghcr.io/dbt-labs/dbt-postgres:latest',
        'docker_platform': 'linux/amd64',
        'docker_volumes': [],
        'docker_env': [],
        'docker_network': 'host',
        'docker_user': None,
        'docker_args': '',
        'shell_path': '/bin/bash',
    }

    def _factory(overrides: dict | None = None) -> RunnerConfig:
        config = default_config.copy()

        if overrides:
            invalid_keys = set(overrides) - set(default_config)
            if invalid_keys:
                raise KeyError(
                    f"Invalid override keys: {invalid_keys}. "
                    f"Allowed keys: {set(default_config)}"
                )

            config.update(cast(RunnerConfig, overrides))

        return config

    return _factory



"""Example end-to-end tests for dbt-ci workflows."""
import pytest
from unittest.mock import patch, MagicMock
from argparse import Namespace


class TestInitWorkflow:
    """Test the init command end-to-end workflow."""
    
    @pytest.fixture
    def mock_args(self):
        """Create mock arguments for init command."""
        return {
            'dbt_project_dir': './dbt',
            'profiles_dir': './dbt',
            'reference_target': 'production',
            'state': './dbt/.dbtstate',
            'runner': 'dbt',
            'dry_run': False,
            'log_level': 'INFO'
        }
    
    def test_init_command_success(self, mock_args):
        """Test successful init command execution."""
        # TODO: Implement once init command structure is finalized
        # When implementing, use patches like:
        # @patch('src.commands.init.run_dbt_command')
        # @patch('src.commands.init.DbtGraph')
        # @patch('src.commands.init.CacheManager')
        pytest.skip("Placeholder test - to be implemented")
    
    def test_init_missing_required_args(self):
        """Test init command fails with missing required arguments."""
        # TODO: Test validation of required arguments
        pytest.skip("Placeholder test - to be implemented")


class TestRunWorkflow:
    """Test the run command end-to-end workflow."""
    
    def test_run_modified_models(self):
        """Test running only modified models."""
        # TODO: Set up test dbt project, modify models, run command
        pytest.skip("Placeholder test - to be implemented")
    
    def test_run_with_defer(self):
        """Test run command with --defer flag."""
        # TODO: Test defer functionality
        pytest.skip("Placeholder test - to be implemented")


# Add more e2e test classes:
# - TestEphemeralWorkflow
# - TestDeleteWorkflow
# - TestCloudStorageIntegration
# - TestDockerRunnerIntegration



import click
from src.cli import common_options, cli
from src.logging import setup_logging
from src.utilities.namespace import to_namespace
from src.commands.ephemeral.index import ephemeral

@cli.command(name='ephemeral')
@common_options
@click.option(
    "--keep-env",
    envvar=['DBT_KEEP_ENV'],
    is_flag=True,
    default=False,
    help="Don't destroy ephemeral environment after run (if supported by runner)"
)
def ephemeral_cmd(**kwargs):
    """Run ephemeral CI check workflow
    
    Uses cached state from 'dbt-ci init' to create and test ephemeral schemas.
    Useful for full integration testing in isolated environments.
    
    Examples:
        # Run after init
        dbt-ci init --state-uri gs://bucket/manifest.json --reference-target production
        dbt-ci ephemeral --runner docker
        
        # Keep environment for debugging
        dbt-ci ephemeral --keep-env
    """
    setup_logging(to_namespace(kwargs).log_level)
    return ephemeral(to_namespace(kwargs, command="ephemeral"))


"""This module contains utility functions for DBT commands, such as compiling the reference state."""
import sys
import logging
from argparse import Namespace
from src.cache import CacheManager
from src.graph.dependency_graph import DbtGraph
from src.utilities.git import GitAdapter
from src.logging import print_exception
from src.schema import RunnerConfig, StateChangeSummary
from src.runners import resolve_dbt_commands, run_dbt_command
from src.utilities.paths import get_manifest_file, get_profile
from src.graph.graph_utils import get_deleted_nodes, get_new_nodes, get_nodes, get_structured_modified_nodes

logger = logging.getLogger(__name__)

def dbt_command_reference_compile(args: Namespace) -> None:
    """Compile the reference DBT project to generate the manifest.json."""
    try:
        command = ["compile"]

        if getattr(args, "skip_reference_compile", False) is True:
            logger.info("Skipping reference compilation as per configuration.")
            return

        # This function can be called separately if users want to compile towards reference state independently, but by default it will be called during init if a different reference target is specified.
        reference_target = getattr(args, "reference_target", None)
        reference_vars = getattr(args, "reference_vars", None)
        if reference_target is None:
            logger.warning("No reference target specified, using current target as reference state for comparison.")
        else:
            command.extend(["--target", reference_target])
            command.extend(["--vars", reference_vars]) if reference_vars else None

        run_dbt_command(
            command_args=resolve_dbt_commands(
                command_args=command, 
                args=args, 
                ignore_keys=["vars", "target"] # Don't pass vars or target when compiling reference manifest
            ),
            runner_config=RunnerConfig(args.__dict__)
        )

        logger.info("DBT project compiled successfully. manifest.json generated.")
    except Exception:
        logger.error("Error during reference compilation", exc_info=True)
        sys.exit(1)

def dbt_command_target_compile(args: Namespace, store_cache: bool = True) -> None:
    """Compile the target DBT project to generate the manifest.json."""
    try:
        cache = CacheManager(args)
        dbt_project_dir = getattr(args, "dbt_project_dir", None)
        target_command = ["compile"]
        target = getattr(args, "target", None)
        reference_target = getattr(args, "reference_target", None)
        is_reference_target_same_as_current = reference_target is None or reference_target == getattr(args, "target", None)

        if dbt_project_dir is None:
            logger.error("dbt_project_dir argument is required for target compilation.")
            sys.exit(1)

        if getattr(args, "skip_target_compile", False) is True:
            logger.info("Skipping target compilation as per configuration.")
            return
        
        if is_reference_target_same_as_current:
            logger.info("Reference target is the same as current target, skipping separate compilation for target state.")
            return

        if target and target != "default":
            target_command.extend(["--target", target])

        run_dbt_command(
            command_args=resolve_dbt_commands(target_command, args),
            runner_config=RunnerConfig(args.__dict__)
        )

        if store_cache:
            target_manifest_file = get_manifest_file(dbt_project_dir)
            cache.write_target_manifest(target_manifest_file)

    except Exception:
        logger.error("Error during target compilation", exc_info=True)
        sys.exit(1)

def dbt_command_state_modified(args: Namespace):
    """Compile the DBT project and return a list of modified nodes compared to the reference state."""
    try:
        git = GitAdapter(args)
        cache = CacheManager(args)
        target_graph = DbtGraph(args)
        reference_graph = DbtGraph(args, is_reference=True)

        changed_files = git.get_changed_files()

        commands = resolve_dbt_commands(["ls", "--select", "state:modified", "--output", "name", "--quiet"], args)
        commands.extend(["--target", getattr(args, "reference_target")])
        commands.extend(["--vars", getattr(args, "reference_vars")]) if getattr(args, "reference_vars", None) else None

        logger.debug(f"Running dbt ls command with arguments: {commands}")

        ls_output = run_dbt_command(
            command_args=commands,
            runner_config=RunnerConfig(args.__dict__)
        )

        if ls_output is None:
            logger.info("No modified nodes found during initialization. Exiting...")
            cache.write_cache()
            sys.exit(0)

        modified_nodes = ls_output.stdout.splitlines()
        target_graph_dict = target_graph.to_dict()
        reference_graph_dict = reference_graph.to_dict()

        new_node_ids = set(get_new_nodes(reference_graph_dict, target_graph_dict) or [])
        deleted_node_ids = set(get_deleted_nodes(reference_graph_dict, target_graph_dict) or [])
        truly_modified_nodes = [n for n in modified_nodes if n not in new_node_ids and n not in deleted_node_ids]

        # Compare against git diff to determine what has been modified vs what is new
        if getattr(args, "no_git", False) is False and len(changed_files.keys()) > 0:
            temp_modified_nodes = get_nodes(
                dependency_graph=target_graph_dict,
                node_ids=truly_modified_nodes
            )

            # We now reference and check against git
            complete_modified_nodes = set()
            for node_id, node_info in temp_modified_nodes.items():
                file_path = node_info['original_file_path']
                if file_path in changed_files["modified"]:
                    complete_modified_nodes.add(node_id)

            truly_modified_nodes = list(complete_modified_nodes)

        state_change_summary: StateChangeSummary = {
            "modified_nodes": get_structured_modified_nodes(get_nodes(
                dependency_graph=target_graph_dict, 
                node_ids=truly_modified_nodes
            )),
            "deleted_nodes": get_structured_modified_nodes(get_nodes(
                dependency_graph=reference_graph_dict, 
                node_ids=list(deleted_node_ids)
            )),
            "new_nodes": get_structured_modified_nodes(get_nodes(
                dependency_graph=target_graph_dict, 
                node_ids=list(new_node_ids)
            ))
        }

        return state_change_summary
    except Exception as e:
        raise Exception(f"Error generating state change summary: {str(e)}")

def clone_command(
    selected_nodes: list[str],
    args: Namespace
) -> None:
    """Helper function to run the dbt command that will create the ephemeral models based on the selected nodes."""
    try:
        profile = get_profile(args)
        threads = profile.get("threads", 5)

        command = resolve_dbt_commands(
            command_args=["clone", "--select", *selected_nodes, "--threads", str(threads)],
            args=args
        )

        logger.debug(f"Running dbt clone command with arguments: {command}")

        run_dbt_command(
            command_args=command,
            runner_config=RunnerConfig(args.__dict__)
        )
    except Exception as e:
        logger.error(f"Error running dbt clone command: {str(e)}")
        print_exception(e)
        sys.exit(1)


"""Module for parsing dbt manifest files and generating dependency graphs."""
import sys
import json
from typing import Literal, Optional, cast
from src.schema import MANIFEST_KEY_MAPPING, DBTManifest, DbtNode, DependencyGraph, DependencyGraphNode, DependencyGraphNodeType, Macro, Node, Source

def skeleton_dependencies_structure():
    """Helper function to create an empty dependencies structure."""
    return {
        "node_dependencies": set(),
        "dependencies_by_type": {
            "model": set(),
            "macro": set(),
            "seed": set(),
            "snapshot": set(),
            "source": set(),
            "test": set(),
            "exposure": set(),
        },
    }

def generate_dependency_graph(manifest_file: DBTManifest) -> DependencyGraph:
    """Generate dependency graph from manifest file.
    Args:
        manifest_file: DBTManifest object representing the manifest file
    """
    child_map = manifest_file.get("child_map", {})
    dependency_graph: DependencyGraph = {
        "metadata": manifest_file.get("metadata", {}),
        "model": {},
        "seed": {},
        "snapshot": {},
        "test": {},
        "macro": {},
        "exposure": {},
        "source": {}
    }

    for key, downstream_dependencies in child_map.items():
        node_type: DependencyGraphNodeType = cast(DependencyGraphNodeType, key.split(".")[0])
        manifest_key = MANIFEST_KEY_MAPPING.get(node_type)
        full_item: DbtNode | None  = manifest_file.get(manifest_key, {}).get(key, None) if manifest_key else None

        # Skip if the node type is not recognized (e.g., "analysis", "docs", etc.)
        if node_type not in dependency_graph.keys():
            continue
        if manifest_key is None:
            print(f"Unknown node type '{node_type}' found in manifest file. Exiting.")
            sys.exit(1)
        if full_item is None:
            print(f"Item '{key}' not found in manifest file under '{manifest_key}'. Skipping.")
            continue
        
        name = full_item.get("name", None)
        compiled_code = full_item.get("compiled_code", None)
        original_file_path = full_item.get("original_file_path", None)
        config = full_item.get("config", {})

        node_type_map: dict[str, set] = {
            "model": set(),
            "macro": set(),
            "seed": set(),
            "snapshot": set(),
            "source": set(),
            "test": set(),
            "exposure": set(),
        }
            
        for dep_id in downstream_dependencies:
            dep_type = dep_id.split(".")[0]
            dep_manifest_key = MANIFEST_KEY_MAPPING.get(dep_type)
            
            if dep_manifest_key and dep_type in node_type_map:
                dep_item = manifest_file.get(dep_manifest_key, {}).get(dep_id, None)
                if dep_item:
                    dep_name = dep_item.get("name")
                    if dep_name:
                        node_type_map[dep_type].add(dep_name)

        dependency_graph[node_type][name] = {
            "name": name,
            "id": key,
            "database": full_item.get("database", None),
            "schema": full_item.get("schema", None),
            "resource_type": full_item.get("resource_type", None),
            "original_file_path": original_file_path,
            "compiled_path": full_item.get("compiled_path", None),
            "compiled_code": compiled_code,
            "config": config,
            "columns": set(full_item.get("columns", {}).keys()),
            "materialized": config.get("materialized", None),
            "incremental_strategy": config.get("incremental_strategy", None),
            "downstream_dependencies": {
                "node_dependencies": set(downstream_dependencies),
                "dependencies_by_type": {
                    "model": node_type_map["model"],
                    "macro": node_type_map["macro"],
                    "seed": node_type_map["seed"],
                    "snapshot": node_type_map["snapshot"],
                    "source": node_type_map["source"],
                    "test": node_type_map["test"],
                    "exposure": node_type_map["exposure"],
                }
            },
            "upstream_dependencies": skeleton_dependencies_structure(),
            "indirect_upstream_dependencies": skeleton_dependencies_structure(),
            "indirect_downstream_dependencies": skeleton_dependencies_structure(),
        }

        #print(full_item.get("depends_on", {}).get("macros", []))
        append_depends_on_nodes(
            dependency_graph=dependency_graph,
            node_type=node_type,
            name=name,
            dependencies=full_item.get("depends_on", {}), 
            manifest_file=manifest_file
        )

    append_upstream_dependencies(dependency_graph, cast(dict, manifest_file))
    append_indirect_dependencies(dependency_graph, "upstream")
    append_indirect_dependencies(dependency_graph, "downstream")

    return dependency_graph

def append_depends_on_nodes(
    dependency_graph: DependencyGraph,
    node_type: DependencyGraphNodeType,
    name: str,
    dependencies: dict[DependencyGraphNodeType, list[str]], 
    manifest_file: DBTManifest
) -> None:
    """Append dependencies from the "depends_on" section of the manifest file to the dependency graph."""
    for dep_type, dep_ids in dependencies.items():
        if dep_ids is None or not isinstance(dep_ids, list):
            continue

        # Add all dep_ids to node_dependencies
        dependency_graph[node_type][name]["upstream_dependencies"]["node_dependencies"].update(dep_ids)
        
        # Retrieve name from manifest
        for dep_id in dep_ids:
            # dep_type from depends_on is "nodes" or "macros"
            # Look up in the correct manifest section
            manifest_section = "macros" if dep_type == "macros" else "nodes"
            node = manifest_file.get(manifest_section, {}).get(dep_id, None)
            
            if node is None:
                continue
                
            node_name = node.get("name")
            if node_name is None:
                continue

            # Map to the correct dependency type
            # For "macros" -> "macro", for "nodes" -> extract from dep_id
            if dep_type == "macros":
                dep_category = "macro"
            else:
                # Extract actual type from dep_id (e.g., "model.project.name" -> "model")
                dep_category = dep_id.split(".")[0]
            
            # Only add if this category is tracked
            if dep_category in dependency_graph[node_type][name]["upstream_dependencies"]["dependencies_by_type"]:
                dependency_graph[node_type][name]["upstream_dependencies"]["dependencies_by_type"][dep_category].add(node_name)

def output_dependency_graph(dependency_graph: DependencyGraph, output_path: str) -> None:
    """Output the dependency graph to a JSON file."""
    with open(output_path, "w", encoding="utf-8") as file:
        json.dump(
            obj=dependency_graph, 
            fp=file, 
            indent=4, 
            default=lambda o: list(o) if isinstance(o, set) else o
        )

def find_node_by_id(dependency_graph: DependencyGraph, node_id: str) -> Optional[DependencyGraphNode]:
    """Find a node in dependency_graph by its ID"""
    node_type = node_id.split(".")[0]
    if node_type in dependency_graph:
        for node_data in dependency_graph[node_type].values():
            if node_data.get("id") == node_id:
                return node_data
    return None


def collect_dependencies_recursively(
    dependency_graph: DependencyGraph, 
    node_data: DependencyGraphNode, 
    visited: set[str], 
    direction: Literal["upstream", "downstream"] = "upstream"
) -> None:
    """Recursively collect upstream or downstream dependencies"""
    dep_key = "upstream_dependencies" if direction == "upstream" else "downstream_dependencies"
    
    for dep_id in node_data[dep_key]["node_dependencies"]:
        if dep_id not in visited:
            visited.add(dep_id)
            dep_node = find_node_by_id(dependency_graph, dep_id)
            if dep_node:
                collect_dependencies_recursively(dependency_graph, dep_node, visited, direction)


def append_upstream_dependencies(dependency_graph: DependencyGraph, manifest_file: dict) -> None:
    """Populate upstream dependencies by reversing downstream dependencies"""
    # Iterate through all nodes and their downstream dependencies
    parent_map = manifest_file.get("parent_map", {})

    for child_id, parent_ids in parent_map.items():
        if len(parent_ids) == 0:
            continue

        child_node_type = child_id.split(".")[0]
        manifest_key = MANIFEST_KEY_MAPPING.get(child_node_type)
        node = manifest_file.get(manifest_key, {}).get(child_id, None).get("name", None)

        if node is None:
            print(f"Node with ID '{child_id}' not found in manifest file under '{manifest_key}'. Skipping.")
            continue

        dependency_graph[child_node_type][node]["upstream_dependencies"]["node_dependencies"].update(parent_ids)

        # Sort by dependency type
        for parent_id in parent_ids:
            parent_node_type = parent_id.split(".")[0]
            manifest_key = MANIFEST_KEY_MAPPING.get(parent_node_type)
            parent_node = manifest_file.get(manifest_key, {}).get(parent_id, None)
            name = parent_node.get("name", None) if parent_node else None

            if name is None:
                print(f"Parent node with ID '{parent_id}' not found in manifest file under '{manifest_key}'. Skipping.")
                continue

            dependency_graph[child_node_type][node]["upstream_dependencies"]["dependencies_by_type"][parent_node_type].add(name)


def append_indirect_dependencies(dependency_graph, direction: Literal["upstream", "downstream"] = "upstream"):
    """Populate indirect dependencies (transitive, excluding direct)
    
    Args:
        dependency_graph: The lineage map to populate
        direction: Either "upstream" or "downstream"
    """
    direct_key = f"{direction}_dependencies"
    indirect_key = f"indirect_{direction}_dependencies"
    
    for node_type, nodes in dependency_graph.items():
        if node_type == "metadata":  # Skip metadata
            continue
        for node_data in nodes.values():
            # Collect all transitive dependencies
            all_indirect = set()
            for direct_dep_id in node_data[direct_key]["node_dependencies"]:
                dep_node = find_node_by_id(dependency_graph, direct_dep_id)
                if dep_node:
                    collect_dependencies_recursively(dependency_graph, dep_node, all_indirect, direction)
            
            # Store indirect dependencies (excluding direct)
            node_data[indirect_key]["node_dependencies"] = all_indirect
            
            # Populate by type
            for indirect_id in all_indirect:
                indirect_node = find_node_by_id(dependency_graph, indirect_id)
                if indirect_node:
                    indirect_type = indirect_id.split(".")[0]
                    # Only add to dependencies_by_type if the indirect_type is tracked
                    if indirect_type in node_data[indirect_key]["dependencies_by_type"]:
                        node_data[indirect_key]["dependencies_by_type"][indirect_type].add(indirect_node["name"])


from src.schema import ConnectorConfig, StorageConnectorConfig
from src.connectors.google.storage import (
    google_storage_client,
    google_download_json,
    google_upload_json
)
from src.connectors.google.bigquery import (
    bigquery_client,
    bigquery_create_datasets,
    bigquery_create_tables,
    bigquery_delete_datasets,
    bigquery_delete_tables,
    bigquery_ephemeral_strategy,
    bigquery_delete_strategy,
    bigquery_migration_strategy,
    bigquery_query
)

bigquery_storage_connector: StorageConnectorConfig = {
    "name": "Google Cloud Storage",
    "client": google_storage_client,
    "upload": google_upload_json,
    "download": google_download_json,
}

bigquery_db_connector: ConnectorConfig = {
    "client": bigquery_client,
    "strategies": {
        "ephemeral": bigquery_ephemeral_strategy,
        "delete": bigquery_delete_strategy,
        "migration": bigquery_migration_strategy
    },
    "methods": {
        "query": bigquery_query,
        "create_datasets": bigquery_create_datasets,
        "delete_datasets": bigquery_delete_datasets,
        "create_tables": bigquery_create_tables,
        "delete_tables": bigquery_delete_tables
    }
}

__init__ = [
    "bigquery_storage_connector",
    "bigquery_db_connector"
]


"""Unit tests for the migration command."""
import pytest
from unittest.mock import MagicMock, patch
from argparse import Namespace
from src.commands.migration.index import migration as index


class TestMigrationCommand:
    """Test the migration command."""
    
    @patch('src.commands.migration.index.get_profile')
    @patch('src.commands.migration.index.DbtGraph')
    @patch('src.commands.migration.index.CacheManager')
    @patch('src.commands.migration.index.logger')
    @patch('src.commands.migration.index.click.secho')
    def test_migration_no_cache(
        self,
        mock_secho,
        mock_logger,
        mock_cache,
        mock_graph,
        mock_get_profile
    ):
        """Test migration command exits when no cache is found."""
        # Setup mocks
        mock_get_profile.return_value = {"type": "bigquery"}
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = None
        mock_cache.return_value = mock_cache_instance
        mock_graph_instance = MagicMock()
        mock_graph.return_value = mock_graph_instance
        
        # Run command - expect SystemExit(1)
        args = Namespace(
            dbt_project_dir='/dbt',
            reference_state='/dbt/.dbtstate',
            dry_run=False
        )
        with pytest.raises(SystemExit) as exc_info:
            index(args)
        
        assert exc_info.value.code == 1
    
    @patch('src.commands.migration.index.get_profile')
    @patch('src.commands.migration.index.CacheManager')
    @patch('src.commands.migration.index.DbtGraph')
    @patch('src.commands.migration.index.get_connector')
    @patch('src.commands.migration.index.logger')
    @patch('src.commands.migration.index.click.secho')
    @patch('src.commands.migration.index.sys.exit')
    def test_migration_unsupported_connector(
        self,
        mock_exit,
        mock_secho,
        mock_logger,
        mock_get_connector,
        mock_graph,
        mock_cache,
        mock_get_profile
    ):
        """Test migration command with unsupported connector."""
        # Setup mocks
        mock_get_profile.return_value = {"type": "bigquery"}
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": {}
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_connector.return_value = {'strategies': {'migration': None}}
        
        # Run command
        args = Namespace(
            dbt_project_dir='/dbt',
            dry_run=False
        )
        index(args)
        
        # Verify error was logged
        mock_logger.error.assert_called_once()
        assert 'does not have a migration strategy' in str(mock_logger.error.call_args)
        
        # Verify exit with error
        mock_exit.assert_called_with(1)
    
    @patch('src.commands.migration.index.get_profile')
    @patch('src.commands.migration.index.CacheManager')
    @patch('src.commands.migration.index.DbtGraph')
    @patch('src.commands.migration.index.get_connector')
    @patch('src.commands.migration.index.get_node_ids_from_structured_nodes')
    @patch('src.commands.migration.index.filter_node_ids_by_multiple_types')
    @patch('src.commands.migration.index.logger')
    @patch('src.commands.migration.index.click.secho')
    @patch('src.commands.migration.index.click.echo')
    def test_migration_no_modified_models(
        self,
        mock_echo,
        mock_secho,
        mock_logger,
        mock_filter,
        mock_get_node_ids,
        mock_get_connector,
        mock_graph,
        mock_cache,
        mock_get_profile
    ):
        """Test migration command when no modified models are found."""
        # Setup mocks
        mock_get_profile.return_value = {"type": "bigquery"}
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": None
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_connector.return_value = {'strategies': {'migration': MagicMock()}}
        mock_get_node_ids.return_value = []
        mock_filter.return_value = []
        
        mock_graph_instance = MagicMock()
        mock_graph.return_value = mock_graph_instance
        
        # Run command - expect SystemExit(0)
        args = Namespace(
            dbt_project_dir='/dbt',
            reference_state='/dbt/.dbtstate',
            dry_run=False
        )
        with pytest.raises(SystemExit) as exc_info:
            index(args)
        
        assert exc_info.value.code == 0
    
    @patch('src.commands.migration.index.get_profile')
    @patch('src.commands.migration.index.CacheManager')
    @patch('src.commands.migration.index.DbtGraph')
    @patch('src.commands.migration.index.get_connector')
    @patch('src.commands.migration.index.get_node_ids_from_structured_nodes')
    @patch('src.commands.migration.index.filter_node_ids_by_multiple_types')
    @patch('src.commands.migration.index.get_nodes')
    @patch('src.commands.migration.index.logger')
    @patch('src.commands.migration.index.click.secho')
    @patch('src.commands.migration.index.click.echo')
    def test_migration_with_modified_models(
        self,
        mock_echo,
        mock_secho,
        mock_logger,
        mock_get_nodes_util,
        mock_filter,
        mock_get_node_ids,
        mock_get_connector,
        mock_graph,
        mock_cache,
        mock_get_profile
    ):
        """Test migration command with modified models."""
        # Setup mocks
        mock_get_profile.return_value = {"type": "bigquery"}
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": {"model": {"model.project.model1": {}}}
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_node_ids.return_value = ['model.project.model1']
        mock_filter.return_value = ['model.project.model1']
        
        # Target node with new partitioning and incremental materialization
        target_nodes = {
            'model.project.model1': {
                'id': 'model.project.model1',
                'resource_type': 'model',
                'name': 'model1',
                'database': 'my_db',
                'schema': 'my_schema',
                'config': {
                    'materialized': 'incremental',
                    'partition_by': {'field': 'new_date', 'data_type': 'date'}
                },
                'compiled_code': 'SELECT * FROM table'
            }
        }
        
        # Reference node with old partitioning
        reference_nodes = {
            'model.project.model1': {
                'id': 'model.project.model1',
                'resource_type': 'model',
                'name': 'model1',
                'database': 'my_db',
                'schema': 'my_schema',
                'config': {
                    'materialized': 'incremental',
                    'partition_by': {'field': 'old_date', 'data_type': 'date'}
                },
                'compiled_code': 'SELECT * FROM table'
            }
        }
        
        # get_nodes is called twice: once for target, once for reference
        mock_get_nodes_util.side_effect = [target_nodes, reference_nodes]
        
        # Graph to_dict returns are not directly used since get_nodes is mocked
        mock_graph_instance = MagicMock()
        mock_graph.return_value = mock_graph_instance
        
        mock_migration_func = MagicMock()
        mock_get_connector.return_value = {'strategies': {'migration': mock_migration_func}}
        
        # Run command - migration should complete successfully without sys.exit()
        args = Namespace(
            dbt_project_dir='/dbt',
            reference_state='/dbt/.dbtstate',
            dry_run=False
        )
        index(args)
        
        # Verify migration function was called with correct migration map
        mock_migration_func.assert_called_once()
        call_args = mock_migration_func.call_args[0]
        migration_map = call_args[0]
        
        # Verify migration map structure
        assert migration_map['connector'] == 'bigquery'
        assert 'model.project.model1' in migration_map['nodes']
        assert migration_map['nodes']['model.project.model1']['table_id'] == 'my_db.my_schema.model1'
        assert migration_map['nodes']['model.project.model1']['old_partitioning'] == {'field': 'old_date', 'data_type': 'date'}
        assert migration_map['nodes']['model.project.model1']['new_partitioning'] == {'field': 'new_date', 'data_type': 'date'}
    
    @patch('src.commands.migration.index.get_profile')
    @patch('src.commands.migration.index.CacheManager')
    @patch('src.commands.migration.index.DbtGraph')
    @patch('src.commands.migration.index.get_connector')
    @patch('src.commands.migration.index.get_node_ids_from_structured_nodes')
    @patch('src.commands.migration.index.filter_node_ids_by_multiple_types')
    @patch('src.commands.migration.index.get_nodes')
    @patch('src.commands.migration.index.logger')
    @patch('src.commands.migration.index.click.secho')
    @patch('src.commands.migration.index.click.echo')
    def test_migration_no_partitioning_changes(
        self,
        mock_echo,
        mock_secho,
        mock_logger,
        mock_get_nodes_util,
        mock_filter,
        mock_get_node_ids,
        mock_get_connector,
        mock_graph,
        mock_cache,
        mock_get_profile
    ):
        """Test migration command when partitioning hasn't changed."""
        # Setup mocks
        mock_get_profile.return_value = {"type": "bigquery"}
        mock_cache_instance = MagicMock()
        mock_cache_instance.get_cache.return_value = {
            "modified_nodes": {"model": {"model.project.model1": {}}}
        }
        mock_cache.return_value = mock_cache_instance
        mock_get_node_ids.return_value = ['model.project.model1']
        mock_filter.return_value = ['model.project.model1']
        
        # Same partitioning in both reference and target
        partitioning = {'field': 'date', 'data_type': 'date'}
        
        mock_get_nodes_util.return_value = {
            'model.project.model1': {
                'id': 'model.project.model1',
                'resource_type': 'model',
                'name': 'model1',
                'config': {'partition_by': partitioning}
            }
        }
        
        mock_graph_instance = MagicMock()
        mock_graph_instance.to_dict.return_value = {
            "model": {
                "model.project.model1": {
                    'config': {'partition_by': partitioning}
                }
            }
        }
        mock_graph.return_value = mock_graph_instance
        
        mock_migration_func = MagicMock()
        mock_get_connector.return_value = {'strategies': {'migration': mock_migration_func}}
        
        # Run command - expect SystemExit(0)
        args = Namespace(
            dbt_project_dir='/dbt',
            reference_state='/dbt/.dbtstate',
            dry_run=False
        )
        with pytest.raises(SystemExit) as exc_info:
            index(args)
        
        assert exc_info.value.code == 0



"""Storage connector for AWS S3 interactions."""
import json
import boto3

from src.schema import DBTManifest

def aws_storage_client():
    """Initialize AWS S3 client."""
    return boto3.client("s3")

def aws_upload_json(uri: str, data: dict) -> None:
    """Upload a JSON object to the specified S3 URI."""
    if not uri.startswith("s3://"):
        raise ValueError(f"Invalid S3 URI: {uri}. Must start with 's3://'.")
    
    # Parse bucket name and key from URI
    path_parts = uri[5:].split("/", 1)
    bucket_name = path_parts[0]
    key = path_parts[1] if len(path_parts) > 1 else ""

    try:
        client = aws_storage_client()
        client.put_object(Bucket=bucket_name, Key=key, Body=json.dumps(data), ContentType="application/json")
        print(f"Successfully uploaded JSON to {uri}")
    except Exception as e:
        raise RuntimeError(f"Failed to upload JSON to {uri}: {e}")
    
def aws_download_json(uri: str) -> DBTManifest:
    """Download and parse a JSON object from the specified S3 URI."""
    if not uri.startswith("s3://"):
        raise ValueError(f"Invalid S3 URI: {uri}. Must start with 's3://'.")
    
    # Parse bucket name and key from URI
    path_parts = uri[5:].split("/", 1)
    bucket_name = path_parts[0]
    key = path_parts[1] if len(path_parts) > 1 else ""

    try:
        client = aws_storage_client()
        response = client.get_object(Bucket=bucket_name, Key=key)
        contents = response['Body'].read().decode('utf-8')
        return json.loads(contents)
    except json.JSONDecodeError as e:
        raise RuntimeError(f"Failed to parse JSON from {uri}: {e}")
    except Exception as e:
        raise RuntimeError(f"Failed to download JSON from {uri}: {e}")


"""
    Ephemeral command for dbt-ci

    Models: By default, select all downstream dependencies (full graph)
    Tests: By default, select upstream depedendenceis (1 level)
    Snapshots: By default, only select upstream dependencies (1 level)
"""

import sys
import logging
from itertools import chain
from argparse import Namespace
from typing import Optional
import click
from src.cache import CacheManager
from src.connectors import DB_CONNECTORS
from src.graph.dependency_graph import DbtGraph
from src.logging import print_exception
from src.schema import EphemeralMapNode
from src.utilities.dbt_commands import clone_command
from src.utilities.paths import get_profile
from src.graph.graph_utils import (
    filter_node_ids_by_multiple_types,
    get_downstream_dependencies,
    get_node_ids_from_structured_nodes,
    get_nodes,
    get_upstream_dependencies
)

logger = logging.getLogger(__name__)

def ephemeral(args: Namespace):
    """Run ephemeral CI check workflow
    
    Detects changes, lists modified models, but doesn't execute them.
    Useful for quick CI checks and PR previews.
    
    Examples:
        dbt-ci ephemeral --state prod-manifest/ --dbt-project-dir ./dbt
        dbt-ci ephemeral --state prod-manifest/ --runner docker
        
        # With environment variables
        export DBT_STATE=./dbt/.dbtstate/
        export DBT_PROJECT_DIR=./dbt
        dbt-ci ephemeral
    """
    try:
        # Convert kwargs to Namespace and resolve configuration
        # Variables class handles type conversions (tuples->lists, string->bool, etc.)
        click.secho("DBT CI Ephemeral", fg="green", bold=True)
        logger.debug(f"Running with the following arguments: {args}")
        cache = CacheManager(args)
        cache.start_report("ephemeral", args)
        connector_type = get_profile(args)["type"]
        
        # Validate connector before calling get_connector
        if connector_type is None:
            logger.error(f"Missing connector type for ephemeral mode: {connector_type}. Supported connectors: {list(DB_CONNECTORS.keys())}")
            sys.exit(1)
        elif connector_type not in DB_CONNECTORS:
            logger.error(f"Unsupported connector type for ephemeral mode: {connector_type}. Supported connectors: {list(DB_CONNECTORS.keys())}")
            sys.exit(1)
        
        #ephemeral_connector = get_connector(connector_type).get("strategies", {}).get("ephemeral")
        ephemeral_map = generate_ephemeral_map(args, cache)

        # Pass the ephemeral map and variables to the connector strategy which
        # will handle the ephemeral execution logic based on the connector type
        if getattr(args, "dry_run", False):
            if getattr(args, "quiet", False):
                sys.exit(0)
            
            logger.info("Dry run mode enabled - no actual ephemeral environment will be created.")
            logger.info("\n------------------------------------------------------")
            for node_id, node_info in ephemeral_map.items():
                target_table_id = f"{node_info['ephemeral_config']['database']}.{node_info['ephemeral_config']['schema']}.{node_info['ephemeral_config']['name']}" if node_info['ephemeral_config'] else "N/A"
                reference_table_id = f"{node_info['reference_config']['database']}.{node_info['reference_config']['schema']}.{node_info['reference_config']['name']}" if node_info['reference_config'] else "N/A"

                logger.info(f"Model: {node_id}")
                logger.info(f"  - Ephemeral Target: {target_table_id}")
                logger.info(f"  - Reference Target: {reference_table_id}") 
            logger.info("------------------------------------------------------\n")
            sys.exit(0)
        elif len(ephemeral_map) == 0:
            logger.info("No nodes found to create ephemeral environment for. Exiting...")
            sys.exit(0)

        #ephemeral_connector(ephemeral_map, args)
        clone_command(list(ephemeral_map.keys()), args)
        
        # Store cache (ephemeral map) for use in finalize step
        cache.write_ephemeral(ephemeral_map)
        cache.update_report("ephemeral", "completed", comment=str(list(ephemeral_map.keys())))
        logger.info("Ephemeral strategy completed successfully.")
        logger.info("Now you can run your dbt command with the appropriate selection to target the ephemeral models and their downstream dependencies.")
        sys.exit(0)
    except Exception as e:
        cache = CacheManager(args)
        cache.update_report("ephemeral", "failed", comment=str(e))
        print_exception(e)
        sys.exit(1)

def generate_ephemeral_map(args: Namespace, cache: CacheManager) -> dict[str, EphemeralMapNode]:
    """Generate a map of nodes to be included in the ephemeral environment based on the cache and dependency graph."""
    target_graph = DbtGraph(args)
    reference_graph = DbtGraph(args, is_reference=True)

    # Look for cache
    prev_cache = cache.get_cache()
    if prev_cache is None: # Should we exit here instead of compiling?
        logger.error("No cache found, please run 'dbt-ci init' first to generate the necessary manifest files and cache for comparison.")
        sys.exit(1)
    logger.info("Cache successfully found - using cached state for comparison")

    changed_nodes_dict = {
        "modified_nodes": get_node_ids_from_structured_nodes(cache.get_cache().get("modified_nodes", None)) or [],
        "deleted_nodes": get_node_ids_from_structured_nodes(cache.get_cache().get("deleted_nodes", None)) or [],
        "new_nodes": get_node_ids_from_structured_nodes(cache.get_cache().get("new_nodes", None)) or [],
    }

    changed_nodes = list(set(chain(
        changed_nodes_dict["modified_nodes"],
        changed_nodes_dict["deleted_nodes"],
        changed_nodes_dict["new_nodes"],
    )))

    if len(changed_nodes) == 0:
        logger.info("No modified, deleted, or new nodes found in cache, skipping...")
        sys.exit(0)

    # Ephemeral cloning strategy:
    # 1. All modified nodes (models, snapshots, tests)
    # 2. Downstream dependencies of modified & deleted nodes
    # 3. If a snapshot changed, include its upstream dependencies (first level)
    # 4. If a test changed, include its upstream dependencies (first level)
    # Note: New nodes are skipped - they should be created in the PR/merge, not cloned
        
    # Get modified nodes of type that needs to have upstream dependencies included
    # Snapshots & tests
    nodes_with_upstream_deps = list(set(chain(
        filter_node_ids_by_multiple_types(
            dependency_graph=target_graph.to_dict(),
            node_types=["snapshot", "test"],
            node_ids=list(chain(
                changed_nodes_dict["modified_nodes"],
                changed_nodes_dict["new_nodes"],
            ))
        )
    )))
        
    selected_nodes = filter_node_ids_by_multiple_types(
        dependency_graph=target_graph.to_dict(),
        node_types=["model", "snapshot"],
        node_ids=list(chain(
            # 1. All modified nodes (models, snapshots)
            changed_nodes_dict["modified_nodes"],
            changed_nodes_dict["new_nodes"],
            # 2. Upstream dependencies of modified models (first level, only models)
            get_upstream_dependencies(target_graph.to_dict(), changed_nodes, "model") or [],
            # 3. Get upstream dependencies of modified models downstream dependencies (full graph, all types)
            get_upstream_dependencies(
                dependency_graph=target_graph.to_dict(), 
                node_ids=list(get_downstream_dependencies(target_graph.to_dict(), changed_nodes, None) or []),
            ) or [],
            # 4. Indirect Downstream dependencies of modified & deleted nodes (all types)
            get_downstream_dependencies(target_graph.to_dict(), changed_nodes, None) or [],
            # 5. Upstream dependencies of modified snapshots & tests (first level, any type)
            get_upstream_dependencies(target_graph.to_dict(), nodes_with_upstream_deps, None) or [],
        ))
    )

    # Lets get all metadata related to these downstream dependencies
    #print(downstream_dependencies)
    target_nodes = get_nodes(target_graph.to_dict(), selected_nodes)
    reference_nodes = get_nodes(reference_graph.to_dict(), selected_nodes)

    # Now we build a dict of target & reference nodes for the engine
    # to decide how to execute and create the ephemeral environment
    ephemeral_map: dict[str, EphemeralMapNode] = {}

    # We use reference since it will also include deleted nodes
    # In target, they wont exist and return None
    for node_id, node_metadata in reference_nodes.items():
        # Skip ephemeral models since they are not materialized in the database
        # and therefore cannot be cloned
        if (
            node_metadata["materialized"] in ("ephemeral", "view", "table")
            and node_metadata["resource_type"] == "model"
        ):
            continue
            
        target_node = target_nodes.get(node_id, None)
        if target_node is None:
            continue

        ephemeral_map[node_id] = {
            "name": node_metadata["name"],
            "resource_type": node_metadata["resource_type"],
            "ephemeral_config": full_config_or_none(
                database=target_node.get("database", None),
                schema=target_node.get("schema", None),
                name=target_node.get("name", None),
                alias=target_node.get("alias", None),
            ),
            "reference_config": full_config_or_none(
                database=node_metadata.get("database", None),
                schema=node_metadata.get("schema", None),
                name=node_metadata.get("name", None),
                alias=node_metadata.get("alias", None),
            )
        }

    return ephemeral_map
    
def full_config_or_none(
    database: Optional[str],
    schema: Optional[str],
    name: Optional[str],
    alias: Optional[str] = None
) -> dict[str, Optional[str]] | None:
    """Helper function to return full config dict if all values are present, otherwise None."""
    if any(v is None for v in (database, schema, name)):
        return None
    return {
        "database": database,
        "schema": schema,
        "name": name,
        "alias": alias,
    }


from src.cli import common_options, cli
from src.utilities.namespace import to_namespace
from src.logging import setup_logging
from src.commands.migration.index import migration

@cli.command(name="migration")
@common_options
def migration_cmd(**kwargs):
    """Run migration check workflow
    
    Uses cached state from 'dbt-ci init' to detect breaking changes and run migration checks.
    Useful for validating changes before merging to production.
    
    Examples:
        # Run after init
        dbt-ci init --state-uri gs://bucket/manifest.json --reference-target production
        dbt-ci migration --runner docker
    """
    setup_logging(to_namespace(kwargs).log_level)
    return migration(to_namespace(kwargs, command="migration"))